(* 'top'-like tool for libvirt domains.
 * $Id: virt-top.ml,v 1.6 2007/08/06 18:43:24 rjones Exp $
 *)

open Printf
open ExtList
open Curses

module C = Libvirt.Connect
module D = Libvirt.Domain
module N = Libvirt.Network

(* To enable debug (to file virt-top.debug) *)
let enable_debug = false

(* When debugging is enabled, it debugs to a file called virt-top.debug. *)
let debug =
  if enable_debug then (
    let chan = open_out "virt-top.debug" in
    fun msg ->
      output_string chan msg;
      output_char chan '\n';
      flush chan
  ) else
    fun _ -> ()

(* Int64 operators for convenience. *)
let (+^) = Int64.add
let (-^) = Int64.sub
let ( *^ ) = Int64.mul
let (/^) = Int64.div

type sort_order = DomainID | DomainName | Processor | Memory | Time

(* Settings. *)
let quit = ref false
let delay = ref 3000 (* milliseconds *)
let historical_cpu_delay = ref 20 (* secs *)
let iterations = ref (-1)
let batch_mode = ref false
let secure_mode = ref false
let sort_order = ref Processor
let uri = ref None

(* Read command line arguments. *)
let () =
  let set_delay newdelay =
    if newdelay <= 0. then
      failwith "-d: cannot set a negative delay";
    delay := int_of_float (newdelay *. 1000.)
  in
  let set_uri u =
    match u with
    | "" -> uri := None
    | _ -> uri := Some u
  in
  let set_sort = function
    | "cpu" | "processor" -> sort_order := Processor
    | "mem" | "memory" -> sort_order := Memory
    | "time" -> sort_order := Time
    | "id" -> sort_order := DomainID
    | "name" -> sort_order := DomainName
    | str -> failwith (str ^ ": sort order should be: cpu|mem|time|id|name")
  in
  let argspec = Arg.align [
    "-b", Arg.Set batch_mode, " Batch mode";
    "-c", Arg.String set_uri, "uri Connect to URI (default: Xen)";
    "--connect", Arg.String set_uri, "uri Connect to URI (default: Xen)";
    "-d", Arg.Float set_delay, "delay Delay time interval (seconds)";
    "-n", Arg.Set_int iterations, "iterations Number of iterations to run";
    "-o", Arg.String set_sort, "sort Set sort order (cpu|mem|time|id|name)";
    "-s", Arg.Set secure_mode, " Secure (\"kiosk\") mode";
    "--hist-cpu", Arg.Set_int historical_cpu_delay, "secs Historical CPU delay";
  ] in
  let anon_fun str = raise (Arg.Bad (str ^ ": unknown parameter")) in
  let usage_msg = "virt-top : a 'top'-like utility for virtualization

SUMMARY
  virt-top [-options]

OPTIONS" in
  Arg.parse argspec anon_fun usage_msg

(* Connect to the hypervisor before going into curses mode, since
 * this is the most likely thing to fail.
 *)
let conn =
  try
    let name = !uri in
    C.connect_readonly ?name ()
  with
    Libvirt.Virterror err ->
      prerr_endline (Libvirt.Virterror.to_string err);
      exit 1

(* Get the node_info.  This never changes, right?  So we get it just once. *)
let node_info = C.get_node_info conn

(* Curses voodoo (see ncurses(3)). *)
let stdscr =
  initscr ();
  cbreak ();
  noecho ();
  nonl ();
  let stdscr = stdscr () in
  intrflush stdscr false;
  keypad stdscr true;
  stdscr

(* Line numbers. *)
let top_lineno = 0
let summary_lineno = 1 (* this takes 2 lines *)
let message_lineno = 3
let header_lineno = 4
let domains_lineno = 5

(* Print in the "message area". *)
let clear_msg () = move message_lineno 0; clrtoeol ()
let print_msg str = clear_msg (); mvaddstr message_lineno 0 str; ()

(* Update the display and sleep for given number of seconds. *)
let sleep n = refresh (); Unix.sleep n

(* The curses getstr/getnstr functions are just weird.
 * This helper function also enables echo temporarily.
 *)
let get_string maxlen =
  echo ();
  let str = String.create maxlen in
  let ok = getstr str in (* Safe because binding calls getnstr. *)
  noecho ();
  if not ok then ""
  else (
    (* Chop at first '\0'. *)
    try
      let i = String.index str '\000' in
      String.sub str 0 i
    with
      Not_found -> str (* it is full maxlen bytes *)
  )

(* Pad a string to the full width with spaces.  If too long, truncate. *)
let pad width str =
  let n = String.length str in
  if n = width then str
  else if n > width then String.sub str 0 width
  else (* if n < width then *) str ^ String.make (width-n) ' '

(* Redraw the display. *)
let redraw =
  (* We save the state of domains across redraws here, which allows us
   * to deduce %CPU usage from the running total.
   *)
  let last_info = Hashtbl.create 13 in
  let last_time = ref (Unix.time ()) in

  (* Keep a historical list of %CPU usages. *)
  let historical_cpu = ref [] in
  let historical_cpu_last_time = ref (Unix.time ()) in

  fun () ->
    clear ();

    (* Get the screen/window size. *)
    let lines, cols = get_size () in

    (* Get the current time. *)
    let time = Unix.time () in
    let tm = Unix.localtime time in
    mvaddstr top_lineno 0
      (sprintf "virt-top %02d:%02d:%02d - "
	 tm.Unix.tm_hour tm.Unix.tm_min tm.Unix.tm_sec);

    (* Basic node_info. *)
    addstr (sprintf "%s %d/%dCPU %dMHz %LdMB "
	      node_info.C.model
	      node_info.C.cpus (C.maxcpus_of_node_info node_info)
	      node_info.C.mhz
	      (node_info.C.memory /^ 1024L));
    (* Save the cursor position for when we come to draw the
     * historical CPU times (down in this function).
     *)
    let historical_cursor = getyx stdscr in

    (* Get the domains.  Match up with their last_info (if any). *)
    let doms =
      (* Active domains. *)
      let n = C.num_of_domains conn in
      let ids = Array.to_list (C.list_domains conn n) in
      let doms =
	List.filter_map (
	  fun id ->
	    try
	      let dom = D.lookup_by_id conn id in
	      let name = D.get_name dom in
	      let info = D.get_info dom in
	      let prev_info =
		try Some (Hashtbl.find last_info id)
		with Not_found -> None in
	      Some (name, Some (id, dom, info, prev_info))
	    with
	      Libvirt.Virterror _ -> None (* ignore transient error *)
	) ids in

      (* Inactive domains. *)
      let n = C.num_of_defined_domains conn in
      let names = Array.to_list (C.list_defined_domains conn n) in
      let doms_inactive = List.map (fun name -> name, None) names in

      doms @ doms_inactive in

    (* What's the total CPU time elapsed since we were last called? (ns) *)
    let total_cpu =
      1_000_000_000. *. float node_info.C.cpus *. (time -. !last_time) in
    (* Avoid division by zero. *)
    let total_cpu = if total_cpu <= 0. then 1. else total_cpu in

    (* Get the CPU time (ns) and %CPU used by each domain. *)
    let doms =
      List.map (
	function
	| name, Some (domid, dom, info, Some prev_info) ->
	    let cpu_time =
	      Int64.to_float (info.D.cpu_time -^ prev_info.D.cpu_time) in
	    let percent_cpu = 100. *. cpu_time /. total_cpu in
	    name, Some (domid, dom, info, Some prev_info, cpu_time, percent_cpu)
	| name, Some (domid, dom, info, None) ->
	    (* No previous history, assume 0. *)
	    name, Some (domid, dom, info, None, 0., 0.)
	| name, None ->
	    name, None
      ) doms in

    (* Sort domains. *)
    let doms =
      let cmp =
	match !sort_order with
	| Processor ->
	    (function
	     | Some (_, _, _, _, _, pc1), Some (_, _, _, _, _, pc2) ->
		 compare pc2 pc1
	     | Some _, None -> -1
	     | None, Some _ -> 1
	     | None, None -> 0)
	| Memory ->
	    (function
	     | Some (_, _, info1, _, _, _), Some (_, _, info2, _, _, _) ->
		 compare info2.D.memory info1.D.memory
	     | Some _, None -> -1
	     | None, Some _ -> 1
	     | None, None -> 0)
	| Time ->
	    (function
	     | Some (_, _, info1, _, _, _), Some (_, _, info2, _, _, _) ->
		 compare info2.D.cpu_time info1.D.cpu_time
	     | Some _, None -> -1
	     | None, Some _ -> 1
	     | None, None -> 0)
	| DomainID ->
	    (function
	     | Some (id1, _, _, _, _, _), Some (id2, _, _, _, _, _) ->
	       compare id1 id2
	     | Some _, None -> -1
	     | None, Some _ -> 1
	     | None, None -> 0)
	| DomainName ->
	    fun _ -> 0 (* fallthrough to default name compare *)
      in
      let cmp (name1, dom1) (name2, dom2) =
	let r = cmp (dom1, dom2) in
	if r <> 0 then r
	else compare name1 name2
      in
      List.sort ~cmp doms in

    (* Show a percentage in 4 chars. *)
    let show_percent percent =
      if percent <= 0. then " 0.0"
      else if percent < 10. then sprintf "% 1.1f" percent
      else if percent < 100. then sprintf "%2.1f" percent
      else "100 "
    in
    (* Format the total time (may be large!) in 9 chars. *)
    let show_time ns =
      let secs_in_ns = 1_000_000_000L in
      let mins_in_ns = 60_000_000_000L in
      let hours_in_ns = 3_600_000_000_000L in

      let hours = ns /^ hours_in_ns in
      let ns = ns -^ (hours *^ hours_in_ns) in
      let mins = ns /^ mins_in_ns in
      let ns = ns -^ (mins *^ mins_in_ns) in
      let secs = ns /^ secs_in_ns in
      let ns = ns -^ (secs *^ secs_in_ns) in
      let pennies = ns /^ 10_000_000L in

      if hours < 12L then
	sprintf "%3Ld:%02Ld.%02Ld" (hours *^ 60L +^ mins) secs pennies
      else if hours <= 999L then
	sprintf "%3Ld:%02Ld:%02Ld" hours mins secs
      else (
	let days = hours /^ 24L in
	let hours = hours -^ (days *^ 24L) in
	sprintf "%3Ldd%02Ld:%02Ld" days hours mins
      )
    in

    (* Print domains. *)
    attron A.reverse;
    mvaddstr header_lineno 0 (pad cols "   ID S %CPU %MEM    TIME   NAME");
    attroff A.reverse;

    let rec loop lineno = function
      | [] -> ()
      | (name, Some (domid, dom, info, _, _, percent_cpu)) :: doms ->
	  if lineno < lines then (
	    let state = match info.D.state with
	      | D.InfoNoState -> '?'
	      | D.InfoRunning -> 'R'
	      | D.InfoBlocked -> 'S'
	      | D.InfoPaused -> 'P'
	      | D.InfoShutdown -> 'D'
	      | D.InfoShutoff -> 'O'
	      | D.InfoCrashed -> 'X'
	    in
	    let percent_cpu = show_percent percent_cpu in
	    let percent_mem = 100L *^ info.D.memory /^ node_info.C.memory in
	    let percent_mem = Int64.to_float percent_mem in
	    let percent_mem = show_percent percent_mem in
	    let time = show_time info.D.cpu_time in
	    let line = sprintf "%5d %c %s %s %s %s"
	      domid state percent_cpu percent_mem time name in
	    let line = pad cols line in
	    mvaddstr lineno 0 line;
	    loop (lineno+1) doms
	  )
      | (name, None) :: doms ->
	  if lineno < lines then (
	    let line = sprintf "    -                       (%s)" name in
	    let line = pad cols line in
	    mvaddstr lineno 0 line;
	    loop (lineno+1) doms
	  )
    in
    loop domains_lineno doms;

    (* Calculate and print totals. *)
    let () =
      let totals = List.fold_left (
	fun (count, running, blocked, paused, shutdown, shutoff,
	     crashed, active, inactive,
	     total_cpu_time, total_memory, total_domU_memory) ->
	  function
	  | (name, Some (domid, _, info, _, cpu_time, _)) ->
	      let test state orig =
		if info.D.state = state then orig+1 else orig
	      in
	      let running = test D.InfoRunning running in
	      let blocked = test D.InfoBlocked blocked in
	      let paused = test D.InfoPaused paused in
	      let shutdown = test D.InfoShutdown shutdown in
	      let shutoff = test D.InfoShutoff shutoff in
	      let crashed = test D.InfoCrashed crashed in

	      let total_cpu_time = total_cpu_time +. cpu_time in
	      let total_memory = total_memory +^ info.D.memory in
	      let total_domU_memory = total_domU_memory +^
		if domid > 0 then info.D.memory else 0L in

	      (count+1, running, blocked, paused, shutdown, shutoff,
	       crashed, active+1, inactive,
	       total_cpu_time, total_memory, total_domU_memory)

	  | (name, None) -> (* inactive domain *)
	      (count+1, running, blocked, paused, shutdown, shutoff,
	       crashed, active, inactive+1,
	       total_cpu_time, total_memory, total_domU_memory)
      ) (0,0,0,0,0,0,0,0,0, 0.,0L,0L) doms in

      let (count, running, blocked, paused, shutdown, shutoff,
	   crashed, active, inactive,
	   total_cpu_time, total_memory, total_domU_memory) = totals in

      mvaddstr summary_lineno 0
	(sprintf "%d domains, %d active, %d running, %d sleeping, %d paused, %d inactive D:%d O:%d X:%d"
	   count active running blocked paused inactive shutdown shutoff
	   crashed);

      (* Total %CPU used, and memory summary. *)
      let percent_cpu = 100. *. total_cpu_time /. total_cpu in
      mvaddstr (summary_lineno+1) 0
	(sprintf "CPU: %2.1f%%  Mem: %Ld MB (%Ld MB by guests)"
	   percent_cpu (total_memory /^ 1024L) (total_domU_memory /^ 1024L));

      (* Time to grab another historical %CPU for the list? *)
      if time >= !historical_cpu_last_time +. float !historical_cpu_delay
      then (
	historical_cpu := percent_cpu :: List.take 10 !historical_cpu;
	historical_cpu_last_time := time
      );

      (* Display historical CPU time. *)
      let () =
	let x, y = historical_cursor in (* Yes, it's a bug in ocaml-curses *)
	let maxwidth = cols - x in
	let line = String.concat " " (List.map show_percent !historical_cpu) in
	let line = pad maxwidth line in
	mvaddstr y x line;
	() in

      ()
    in

    (* Update last_info, last_time. *)
    last_time := time;
    Hashtbl.clear last_info;
    List.iter (
      function
      | (_, Some (domid, _, info, _, _, _)) ->
	Hashtbl.add last_info domid info
      | _ -> ()
    ) doms

(* Main loop. *)
let rec main_loop () =
  while not !quit do
    redraw ();
    refresh ();

    (* Clear up unused virDomainPtr objects. *)
    Gc.compact ();

    if not !batch_mode then
      get_key_press ()
    else (* Batch mode - just sleep, ignore keys. *)
      Unix.sleep (!delay / 1000);

    (* Max iterations? *)
    if !iterations >= 0 then (
      decr iterations;
      if !iterations = 0 then quit := true
    );
  done

and get_key_press () =
  (* Read the next key, waiting up to !delay milliseconds. *)
  timeout !delay;
  let k = getch () in
  timeout (-1); (* Reset to blocking mode. *)
  if k >= 0 && k <> Char.code ' ' then (
    (* No key or space, just continue, otherwise ... *)
    if k = Char.code 'q' then quit := true
    else if k = Char.code 'h' then show_help ()
    else if k = Char.code 's' || k = Char.code 'd' then change_delay ()
    else if k = Char.code 'M' then sort_order := Memory
    else if k = Char.code 'P' then sort_order := Processor
    else if k = Char.code 'T' then sort_order := Time
    else if k = Char.code 'N' then sort_order := DomainID
    else unknown_command k
  )

and change_delay () =
  print_msg (sprintf "Change delay from %.1f to: " (float !delay /. 1000.));
  let str = get_string 16 in
  (* Try to parse the number. *)
  let error =
    try
      let newdelay = float_of_string str in
      if newdelay <= 0. then (
	print_msg "Delay must be > 0"; true
      ) else (
	delay := int_of_float (newdelay *. 1000.); false
      )
    with
      Failure "float_of_string" ->
	print_msg "Not a valid number"; true in
  sleep (if error then 2 else 1)

and show_help () =
  print_msg "Help not implemented!";
  sleep 2

and unknown_command k =
  print_msg "Unknown command - try 'h' for help";
  sleep 1

(* Make sure we catch any exceptions and clean up the display. *)
let () =
  try
    main_loop ();
    endwin ()
  with
  | Libvirt.Virterror err ->
      endwin ();
      prerr_endline (Libvirt.Virterror.to_string err);
      exit 1
  | exn ->
      endwin ();
      prerr_endline ("Error: " ^ Printexc.to_string exn);
      exit 1
